"""IFD ASCII tag structure parser module."""

import warnings

from exif._ifd_tag._base import Base as BaseIfdTag


class Ascii(BaseIfdTag):

    """IFD ASCII tag structure parser class."""

    def modify(self, value):
        """Modify tag value.

        :param value: new tag value
        :type value: corresponding Python type

        """
        if len(value) > self.count - 1:
            raise ValueError("string must be no longer than original")

        if self.count <= 4:
            self.parent_segment_hex.modify_hex(self.value_offset_addr, value.encode("ascii").hex() + "00")
        else:
            cursor = 0xA + self.value_offset

            for character in value:
                self.parent_segment_hex.modify_hex(cursor, hex(ord(character)).lstrip('0x'))
                cursor += 1
            for _i in range(self.count - 1 - len(value)):
                self.parent_segment_hex.modify_hex(cursor, '00')
                cursor += 1

        self.count = len(value) + 1

    def read(self):
        """Read tag value.

        In string types, the count refers to how many characters exist in the string.

        :returns: tag value
        :rtype: corresponding Python type

        """
        retval_chars = []

        if self.count <= 4:
            for character in range(self.count - 1):  # subtract 1 to ignore null end-of-string
                current_ascii_val = self.parent_segment_hex.read(self.value_offset_addr + character, 1)
                retval_chars.append(chr(int(current_ascii_val, 16)))

        else:  # value does not fit in 4 bytes; rather, it's a pointer
            cursor = 0xA + self.value_offset

            # Subtract 1 from IFD member count to ignore null terminator character.
            for member_index in range(self.count - 1):  # pylint: disable=unused-variable
                current_ascii_val = self.parent_segment_hex.read(cursor, 1)

                if not current_ascii_val:  # pragma: no cover
                    warnings.warn("reached end of string prematurely", RuntimeWarning)
                    break

                retval_chars.append(chr(int(current_ascii_val, 16)))
                cursor += 1

        return ''.join(retval_chars)



"""Test deleting EXIF attributes."""

import os
import textwrap
import unittest

from exif import Image
from exif.tests.delete_exif_baselines import (
    DELETE_ASCII_TAGS_HEX_BASELINE, DELETE_GEOTAG_HEX_BASELINE)

# pylint: disable=pointless-statement, protected-access


class TestModifyExif(unittest.TestCase):

    """Test cases for deleting EXIF attributes."""

    def setUp(self):
        """Open sample image file in binary mode for use in test cases."""
        grand_canyon = os.path.join(os.path.dirname(__file__), 'grand_canyon.jpg')
        with open(grand_canyon, 'rb') as image_file:
            self.image = Image(image_file)

        assert self.image.has_exif

    def test_delete_ascii_tags(self):
        """Verify deleting EXIF ASCII from the Image object and the hexadecimal equivalent."""
        del self.image.make
        del self.image.model

        with self.assertRaisesRegex(AttributeError, "image does not have attribute make"):
            self.image.make
        with self.assertRaisesRegex(AttributeError, "image does not have attribute model"):
            self.image.model

        segment_hex = self.image._segments['APP1'].get_segment_hex()
        self.assertEqual('\n'.join(textwrap.wrap(segment_hex, 90)), DELETE_ASCII_TAGS_HEX_BASELINE)

    def test_delete_gps_tags(self):
        """Verify deleting EXIF geotags from the Image object and the hexadecimal equivalent."""
        del self.image.gps_latitude
        del self.image.gps_longitude
        del self.image.gps_altitude

        with self.assertRaisesRegex(AttributeError, "image does not have attribute gps_latitude"):
            self.image.gps_latitude
        with self.assertRaisesRegex(AttributeError, "image does not have attribute gps_longitude"):
            self.image.gps_longitude
        with self.assertRaisesRegex(AttributeError, "image does not have attribute gps_altitude"):
            self.image.gps_altitude

        segment_hex = self.image._segments['APP1'].get_segment_hex()
        self.assertEqual('\n'.join(textwrap.wrap(segment_hex, 90)), DELETE_GEOTAG_HEX_BASELINE)

    def test_delete_method(self):
        """Test behavior when setting tags using the ``delete()`` method."""
        self.image.delete("model")

        with self.assertRaisesRegex(AttributeError, "image does not have attribute model"):
            self.image.model

    def test_handle_unset_attribute(self):
        """Verify that accessing an attribute not present in an image raises an AttributeError."""
        with self.assertRaisesRegex(AttributeError, "image does not have attribute light_source"):
            del self.image.light_source

    def test_index_deleter(self):
        """Test deleting attributes using index syntax."""
        del self.image["model"]

        with self.assertRaisesRegex(AttributeError, "image does not have attribute model"):
            self.image.model

    def test_standard_delete(self):
        """Verify that writing and deleting non-EXIF attributes behave normally."""
        self.image.dummy_attr = 123
        assert self.image.dummy_attr == 123
        del self.image.dummy_attr
        with self.assertRaisesRegex(AttributeError, "unknown image attribute dummy_attr"):
            self.image.dummy_attr



"""Base IFD tag structure parser module."""

from exif._constants import EXIF_LITTLE_ENDIAN_HEADER, EXIF_POINTER_TAG_ID, GPS_POINTER_TAG_ID


class Base:

    """Base IFD tag structure parser class."""

    def __init__(self, tag, count, value_offset, section_start_address, parent_segment_hex,
                 value_offset_addr):
        self.tag = int(tag, 16)
        self.count = int(count, 16)
        self.value_offset = int(value_offset, 16)
        self.section_start_address = section_start_address
        self.parent_segment_hex = parent_segment_hex
        self.value_offset_addr = value_offset_addr

        if self.parent_segment_hex.endianness == EXIF_LITTLE_ENDIAN_HEADER:
            self.struct_index = -1
        else:
            self.struct_index = 0

    def __eq__(self, other):
        return self.tag == other.tag

    def __repr__(self):
        return "_ifd_tag.Base(tag={}, count={}, value_offset={}, section_start_address={})".format(
            self.tag, self.count, self.value_offset, self.section_start_address)

    def is_exif_pointer(self):
        """Determine if this IFD tag is an EXIF pointer.

        :returns: is EXIF pointer
        :rtype: bool

        """
        return self.tag == EXIF_POINTER_TAG_ID

    def is_gps_pointer(self):
        """Determine if this IFD tag is an GPS pointer.

        :returns: is GPS pointer
        :rtype: bool

        """
        return self.tag == GPS_POINTER_TAG_ID

    def modify(self, value):  # pragma: no cover
        """Modify tag value.

        :param value: new tag value
        :type value: corresponding Python type

        """
        raise NotImplementedError("cannot modify a base/unknown IFD tag instance")

    def read(self):  # pragma: no cover
        """Read tag value.

        :returns: tag value
        :rtype: corresponding Python type

        """
        raise NotImplementedError("cannot read a base/unknown IFD tag instance")



"""Test modifying EXIF attributes."""

import os
import textwrap
import unittest

from baseline import Baseline

from exif import Image
from exif.tests.modify_exif_baselines import (
    MODIFY_ASCII_SAME_LEN_HEX_BASELINE, MODIFY_ASCII_SHORTER_HEX_BASELINE,
    MODIFY_RATIONAL_HEX_BASELINE, MODIFY_SRATIONAL_HEX_BASELINE, ROTATED_GRAND_CANYON_HEX)

# pylint: disable=protected-access


class TestModifyExif(unittest.TestCase):

    """Test cases for modifying EXIF attributes."""

    def setUp(self):
        """Open sample image file in binary mode for use in test cases."""
        grand_canyon = os.path.join(os.path.dirname(__file__), 'grand_canyon.jpg')
        with open(grand_canyon, 'rb') as image_file:
            self.image = Image(image_file)

        assert self.image.has_exif

    def test_handle_unset_attribute(self):
        """Verify that accessing an attribute not present in an image raises an AttributeError."""
        with self.assertRaisesRegex(AttributeError, "image does not have attribute light_source"):
            self.image.light_source = 0

    def test_handle_ascii_too_long(self):
        """Verify that writing a longer string to an ASCII tag raises a ValueError."""
        with self.assertRaisesRegex(ValueError, "string must be no longer than original"):
            self.image.model = "MyArtificiallySetCameraAttribute"

    def test_index_modifier(self):
        """Test modifying attributes using index syntax."""
        self.image["model"] = "MyCamera"
        self.assertEqual(self.image.model, Baseline("""MyCamera"""))

    def test_modify_ascii_same_len(self):
        """Verify that writing a same length string to an ASCII tag updates the tag."""
        self.image.model = "MyCamera"
        self.assertEqual(self.image.model, Baseline("""MyCamera"""))

        segment_hex = self.image._segments['APP1'].get_segment_hex()
        self.assertEqual('\n'.join(textwrap.wrap(segment_hex, 90)),
                         MODIFY_ASCII_SAME_LEN_HEX_BASELINE)

    def test_modify_ascii_shorter(self):
        """Verify that writing a shorter string to an ASCII tag updates the tag."""
        self.image.model = "MyCam"
        self.assertEqual(self.image.model, Baseline("""MyCam"""))

        segment_hex = self.image._segments['APP1'].get_segment_hex()
        self.assertEqual('\n'.join(textwrap.wrap(segment_hex, 90)),
                         MODIFY_ASCII_SHORTER_HEX_BASELINE)

    def test_modify_orientation(self):
        """Verify that modifying the orientation (a short tag) updates the tag value as expected."""
        assert self.image.orientation == 1
        assert repr(self.image.orientation) == Baseline("""<Orientation.TOP_LEFT: 1>""")

        self.image.orientation = 6
        assert self.image.orientation == 6

        segment_hex = self.image._segments['APP1'].get_segment_hex()
        self.assertEqual('\n'.join(textwrap.wrap(segment_hex, 90)),
                         ROTATED_GRAND_CANYON_HEX)

    def test_modify_rational(self):
        """Verify that modifying RATIONAL tags updates the tag values as expected."""
        self.image.gps_altitude = 123.456789
        self.assertEqual(str(self.image.gps_altitude), Baseline("""123.456789"""))
        self.image.gps_latitude = (41.0, 36.0, 33.786)
        self.assertEqual(str(self.image.gps_latitude), Baseline("""(41.0, 36.0, 33.786)"""))

        segment_hex = self.image._segments['APP1'].get_segment_hex()
        self.assertEqual('\n'.join(textwrap.wrap(segment_hex, 90)),
                         MODIFY_RATIONAL_HEX_BASELINE)

    def test_modify_srational(self):
        """Verify that modifying a SRATIONAL tag updates the tag value as expected."""
        self.image.brightness_value = -2.468
        self.assertEqual(str(self.image.brightness_value), Baseline("""-2.468"""))

        segment_hex = self.image._segments['APP1'].get_segment_hex()
        self.assertEqual('\n'.join(textwrap.wrap(segment_hex, 90)),
                         MODIFY_SRATIONAL_HEX_BASELINE)

    def test_set_method(self):
        """Test behavior when setting tags using the ``set()`` method."""
        self.image.set("model", "MyCamera")
        self.assertEqual(self.image.model, Baseline("""MyCamera"""))



"""Regression test for miscellaneous GitHub issues.."""

import os
import unittest

from baseline import Baseline

from exif import Image

# pylint: disable=pointless-statement


class TestGitHubIssue12(unittest.TestCase):

    """Regression test for GitHub issue 12."""

    def setUp(self):
        """Open sample image file in binary mode for use in test cases."""
        test_image_path = os.path.join(os.path.dirname(__file__), 'github_issue_12.jpg')
        with open(test_image_path, 'rb') as image_file:
            self.image = Image(image_file)

        assert self.image.has_exif

    def test_get_method(self):
        """Test reading ASCII tags whose values fit inside the value offset field such that it's not a pointer."""
        self.assertEqual(self.image.image_description, Baseline("""dav"""))
        self.image.image_description = "new"
        self.assertEqual(self.image.image_description, Baseline("""new"""))



"""Test modifying EXIF attributes and getting new file contents."""

import binascii
import os
import sys
import textwrap
import unittest

from exif import Image
from exif.tests.get_file_baselines import MODIFIED_NOISE_FILE_HEX_BASELINE


class TestGetFile(unittest.TestCase):

    """Test cases for modifying EXIF attributes and getting new file contents."""

    def setUp(self):
        """Open sample image file in binary mode for use in test cases."""
        noise = os.path.join(os.path.dirname(__file__), 'noise.jpg')
        with open(noise, 'rb') as image_file:
            self.image = Image(image_file)

        assert self.image.has_exif

    def test_get_file(self):
        """Verify that an image is writable to a file after modifying its EXIF metadata.

        Assert the produced file is equivalent to a known baseline.

        """
        self.image.software = "Python"
        file_hex = binascii.hexlify(self.image.get_file())
        if sys.version_info[0] == 3:
            file_hex = file_hex.decode("utf8")
        self.assertEqual('\n'.join(textwrap.wrap(file_hex, 90)), MODIFIED_NOISE_FILE_HEX_BASELINE)



"""APP1 metadata interface module for EXIF tags."""

from exif._constants import (
    ATTRIBUTE_ID_MAP, ATTRIBUTE_NAME_MAP, BYTES_PER_IFD_TAG_COUNT, BYTES_PER_IFD_TAG_ID,
    BYTES_PER_IFD_TAG_TYPE, BYTES_PER_IFD_TAG_VALUE_OFFSET, BYTES_PER_IFD_TAG_TOTAL,
    ERROR_IMG_NO_ATTR, ExifTypes, USER_COMMENT_CHARACTER_CODE_LEN_BYTES)
from exif._hex_interface import HexInterface

from exif._ifd_tag._ascii import Ascii
from exif._ifd_tag._base import Base as BaseIfdTag
from exif._ifd_tag._byte import Byte
from exif._ifd_tag._long import Long
from exif._ifd_tag._rational import Rational
from exif._ifd_tag._short import Short
from exif._ifd_tag._slong import Slong
from exif._ifd_tag._srational import Srational
from exif._ifd_tag._windows_xp import WindowsXp


class App1MetaData:

    """APP1 metadata interface class for EXIF tags."""

    def _delete_ifd_tag(self, ifd_tag):
        # Get the tag count in the deletion target's IFD section.
        delete_target = self.ifd_tags[ifd_tag.tag]
        section_start_address = delete_target.section_start_address
        section_tag_count = int(self._segment_hex.read(section_start_address, 2), 16)

        # Decrease the tag count by 1 in the file hexadecimal.
        self._segment_hex.modify_number(section_start_address, 2, section_tag_count - 1)

        # Parse over the deletion target's IFD section and remove the tag.
        cursor = section_start_address + 2
        for tag_index in range(section_tag_count):  # pylint: disable=unused-variable
            tag_id = self._segment_hex.read(cursor, BYTES_PER_IFD_TAG_ID)
            cursor += BYTES_PER_IFD_TAG_ID
            tag_type = int(self._segment_hex.read(cursor, BYTES_PER_IFD_TAG_ID), 16)
            cursor += BYTES_PER_IFD_TAG_TYPE
            tag_count = self._segment_hex.read(cursor, BYTES_PER_IFD_TAG_COUNT)
            cursor += BYTES_PER_IFD_TAG_COUNT
            tag_value_offset = self._segment_hex.read(cursor, BYTES_PER_IFD_TAG_VALUE_OFFSET)
            tag_value_offset_addr = cursor
            cursor += BYTES_PER_IFD_TAG_VALUE_OFFSET

            tag = self._tag_factory(tag_id, tag_type, tag_count, tag_value_offset,
                                    section_start_address, tag_value_offset_addr)

            if delete_target == tag:
                cursor -= BYTES_PER_IFD_TAG_TOTAL
                self._segment_hex.delete(cursor, BYTES_PER_IFD_TAG_TOTAL)

        # Pad ending of IFD section to preserve pointers.
        self._segment_hex.insert_null(cursor, BYTES_PER_IFD_TAG_TOTAL)

        # Overwrite pointer data with null bytes (if applicable, depending on datatype).
        cursor = 0xA + ifd_tag.value_offset
        if isinstance(delete_target, Ascii):
            self._segment_hex.wipe(cursor, delete_target.count)
        if isinstance(delete_target, (Srational, Rational)):  # 8 Bytes
            self._segment_hex.wipe(cursor, delete_target.count * 8)

        # Remove tag from parser tag dictionary.
        del self.ifd_tags[ifd_tag.tag]

    def get_segment_hex(self):
        """Get equivalent APP1 hexadecimal string.

        :returns: segment hexadecimal string
        :rtype: str

        """
        return self._segment_hex.get_hex_string()

    def get_tag_list(self):
        """Get a list of EXIF tag attributes present in the image objec.

        :returns: image EXIF tag names
        :rtype: list of str

        """
        return [ATTRIBUTE_NAME_MAP.get(key, "<unknown EXIF tag {0}>".format(key))
                for key in self.ifd_tags]

    def _tag_factory(self, tag_id, tag_type, tag_count, tag_value_offset, section_start_address,
                     tag_value_offset_addr):
        if ATTRIBUTE_ID_MAP["xp_title"] <= int(tag_id, 16) <= ATTRIBUTE_ID_MAP["xp_subject"]:  # legacy Windows XP tags
            tag = WindowsXp(tag_id, tag_count, tag_value_offset,
                            section_start_address, self._segment_hex, tag_value_offset_addr)
        elif tag_type == ExifTypes.BYTE:
            tag = Byte(tag_id, tag_count, tag_value_offset,
                       section_start_address, self._segment_hex, tag_value_offset_addr)
        elif tag_type == ExifTypes.ASCII:
            tag = Ascii(tag_id, tag_count, tag_value_offset,
                        section_start_address, self._segment_hex, tag_value_offset_addr)
        elif tag_type == ExifTypes.SHORT:
            tag = Short(tag_id, tag_count, tag_value_offset,
                        section_start_address, self._segment_hex, tag_value_offset_addr)
        elif tag_type == ExifTypes.LONG:
            tag = Long(tag_id, tag_count, tag_value_offset,
                       section_start_address, self._segment_hex, tag_value_offset_addr)
        elif tag_type == ExifTypes.RATIONAL:
            tag = Rational(tag_id, tag_count, tag_value_offset,
                           section_start_address, self._segment_hex, tag_value_offset_addr)
        elif tag_type == ExifTypes.SLONG:
            tag = Slong(tag_id, tag_count, tag_value_offset,
                        section_start_address, self._segment_hex, tag_value_offset_addr)
        elif tag_type == ExifTypes.SRATIONAL:
            tag = Srational(tag_id, tag_count, tag_value_offset,
                            section_start_address, self._segment_hex, tag_value_offset_addr)
        else:
            tag = BaseIfdTag(tag_id, tag_count, tag_value_offset,
                             section_start_address, self._segment_hex, tag_value_offset_addr)

        return tag

    def _unpack_ifd_tags(self):
        cursor = 0xA
        exif_offset = None
        gps_offset = None

        # Read the endianness specified by the image.
        self._segment_hex.set_endianness(self._segment_hex.read(cursor, 2))
        cursor += 2

        # Skip over fixed 0x002A bytes.
        cursor += 2

        # Determine the location of the first IFD section relative to the start of the APP1 section.
        # 0xA is IFD section offset from start of APP1.
        cursor = 0xA + int(self._segment_hex.read(cursor, 4), 16)

        # Read each IFD section.
        current_ifd = 0
        while cursor != 0xA:
            section_start_address = cursor
            num_ifd_tags = int(self._segment_hex.read(cursor, 2), 16)
            cursor += 2

            for tag_index in range(num_ifd_tags):  # pylint: disable=unused-variable
                tag_id = self._segment_hex.read(cursor, BYTES_PER_IFD_TAG_ID)
                cursor += BYTES_PER_IFD_TAG_ID
                tag_type = int(self._segment_hex.read(cursor, BYTES_PER_IFD_TAG_TYPE), 16)
                cursor += BYTES_PER_IFD_TAG_TYPE
                tag_count = self._segment_hex.read(cursor, BYTES_PER_IFD_TAG_COUNT)
                cursor += BYTES_PER_IFD_TAG_COUNT
                tag_value_offset = self._segment_hex.read(cursor, BYTES_PER_IFD_TAG_VALUE_OFFSET)
                tag_value_offset_addr = cursor
                cursor += BYTES_PER_IFD_TAG_VALUE_OFFSET

                tag = self._tag_factory(tag_id, tag_type, tag_count, tag_value_offset,
                                        section_start_address, tag_value_offset_addr)

                self.ifd_tags[tag.tag] = tag

                if tag.is_exif_pointer():
                    exif_offset = tag.value_offset
                if tag.is_gps_pointer():
                    gps_offset = tag.value_offset

            next_offset_str = self._segment_hex.read(cursor, 4)

            try:
                next_offset_int = int(next_offset_str, 16)
            except ValueError:
                # Handle case of invalid literal for int() with base 16: ''
                next_offset_int = 0

            cursor = 0xA + next_offset_int
            current_ifd += 1

        # If an EXIF section pointer exists, read EXIF IFD tags.
        if exif_offset:
            cursor = 0xA + exif_offset
            section_start_address = cursor
            num_ifd_tags = int(self._segment_hex.read(cursor, 2), 16)
            cursor += 2

            for tag_index in range(num_ifd_tags):  # pylint: disable=unused-variable
                tag_id = self._segment_hex.read(cursor, BYTES_PER_IFD_TAG_ID)
                cursor += BYTES_PER_IFD_TAG_ID
                tag_type = int(self._segment_hex.read(cursor, BYTES_PER_IFD_TAG_ID), 16)
                cursor += BYTES_PER_IFD_TAG_TYPE
                tag_count = self._segment_hex.read(cursor, BYTES_PER_IFD_TAG_COUNT)
                cursor += BYTES_PER_IFD_TAG_COUNT
                tag_value_offset = self._segment_hex.read(cursor, BYTES_PER_IFD_TAG_VALUE_OFFSET)
                tag_value_offset_addr = cursor
                cursor += BYTES_PER_IFD_TAG_VALUE_OFFSET

                tag = self._tag_factory(tag_id, tag_type, tag_count, tag_value_offset,
                                        section_start_address, tag_value_offset_addr)

                # Handle user comment data structure (see pg. 51 of EXIF specification).
                if tag.tag == ATTRIBUTE_ID_MAP["user_comment"]:
                    tag = Ascii(tag_id, tag_count, tag_value_offset,
                                section_start_address, self._segment_hex, tag_value_offset_addr)
                    tag.dtype = ExifTypes.ASCII  # reads unicode as well

                    tag.count += 1  # custom data structure does not use null terminator
                    tag.count -= USER_COMMENT_CHARACTER_CODE_LEN_BYTES
                    tag.value_offset += USER_COMMENT_CHARACTER_CODE_LEN_BYTES

                self.ifd_tags[tag.tag] = tag

        # If an GPS section pointer exists, read GPS IFD tags.
        if gps_offset:
            cursor = 0xA + gps_offset
            section_start_address = cursor
            num_ifd_tags = int(self._segment_hex.read(cursor, 2), 16)
            cursor += 2

            for tag_index in range(num_ifd_tags):  # pylint: disable=unused-variable
                tag_id = self._segment_hex.read(cursor, BYTES_PER_IFD_TAG_ID)
                cursor += BYTES_PER_IFD_TAG_ID
                tag_type = int(self._segment_hex.read(cursor, BYTES_PER_IFD_TAG_ID), 16)
                cursor += BYTES_PER_IFD_TAG_TYPE
                tag_count = self._segment_hex.read(cursor, BYTES_PER_IFD_TAG_COUNT)
                cursor += BYTES_PER_IFD_TAG_COUNT
                tag_value_offset = self._segment_hex.read(cursor, BYTES_PER_IFD_TAG_VALUE_OFFSET)
                tag_value_offset_addr = cursor
                cursor += BYTES_PER_IFD_TAG_VALUE_OFFSET

                tag = self._tag_factory(tag_id, tag_type, tag_count, tag_value_offset,
                                        section_start_address, tag_value_offset_addr)

                self.ifd_tags[tag.tag] = tag

    def __init__(self, segment_hex):
        self._segment_hex = HexInterface(segment_hex)
        self.ifd_tags = {}

        self._unpack_ifd_tags()

    def __delattr__(self, item):
        try:
            # Determine if attribute is an IFD tag accessor.
            attribute_id = ATTRIBUTE_ID_MAP[item]
        except KeyError:  # pragma: no cover
            # Coverage and behavior tested by Image class.
            # Attribute is a class member. Delete natively.
            super(App1MetaData, self).__delattr__(item)
        else:
            # Attribute is not a class member. Delete EXIF tag value.
            try:
                ifd_tag = self.ifd_tags[attribute_id]
            except KeyError:
                raise AttributeError(ERROR_IMG_NO_ATTR.format(item))

            self._delete_ifd_tag(ifd_tag)

    def __getattr__(self, item):
        """If attribute is not a class member, get the value of the EXIF tag of the same name."""
        try:
            attribute_id = ATTRIBUTE_ID_MAP[item]
        except KeyError:
            raise AttributeError("unknown image attribute {0}".format(item))

        try:
            ifd_tag = self.ifd_tags[attribute_id]
        except KeyError:
            raise AttributeError(ERROR_IMG_NO_ATTR.format(item))

        return ifd_tag.read()

    def __setattr__(self, key, value):
        try:
            # Determine if attribute is an IFD tag accessor.
            attribute_id = ATTRIBUTE_ID_MAP[key]
        except KeyError:
            # Attribute is a class member. Set natively.
            super(App1MetaData, self).__setattr__(key, value)
        else:
            try:
                ifd_tag = self.ifd_tags[attribute_id]
            except KeyError:
                raise AttributeError(ERROR_IMG_NO_ATTR.format(key))

            ifd_tag.modify(value)



"""Test reading EXIF attributes."""

import os
import unittest

from baseline import Baseline

from exif import Image

# pylint: disable=pointless-statement


class TestReadExif(unittest.TestCase):

    """Test cases for reading EXIF attributes."""

    def setUp(self):
        """Open sample image file in binary mode for use in test cases."""
        grand_canyon = os.path.join(os.path.dirname(__file__), 'grand_canyon.jpg')
        with open(grand_canyon, 'rb') as image_file:
            self.image = Image(image_file)

        assert self.image.has_exif

    def test_get_method(self):
        """Test behavior when accessing tags using the ``get()`` method."""
        self.assertIsNone(self.image.get('fake_attribute'))
        self.assertEqual(self.image.get('light_source', default=-1), -1)  # tag not in image
        self.assertEqual(self.image.get('make'), Baseline("""Apple"""))

    def test_handle_bad_attribute(self):
        """Verify that accessing a nonexistent attribute raises an AttributeError."""
        with self.assertRaisesRegex(AttributeError, "unknown image attribute fake_attribute"):
            self.image.fake_attribute

    def test_handle_unset_attribute(self):
        """Verify that accessing an attribute not present in an image raises an AttributeError."""
        with self.assertRaisesRegex(AttributeError, "image does not have attribute light_source"):
            self.image.light_source

    def test_index_accessor(self):
        """Test accessing attributes using index syntax."""
        self.assertEqual(self.image["datetime"], Baseline("""2018:03:12 10:12:07"""))

    def test_read_ascii(self):
        """Test reading ASCII tags and compare to known baseline values."""
        self.assertEqual(self.image.datetime, Baseline("""2018:03:12 10:12:07"""))
        self.assertEqual(self.image.make, Baseline("""Apple"""))
        self.assertEqual(self.image.model, Baseline("""iPhone 7"""))
        self.assertEqual(self.image.gps_latitude_ref, Baseline("""N"""))
        self.assertEqual(self.image.gps_longitude_ref, Baseline("""W"""))

    def test_read_byte(self):
        """Test reading BYTE tags and compare to known baseline values."""
        self.assertEqual(str(self.image.gps_altitude_ref), Baseline("""0"""))

    def test_read_long(self):
        """Test reading LONG tags and compare to known baseline values."""
        self.assertEqual(str(self.image.jpeg_interchange_format), Baseline("""6410"""))
        self.assertEqual(str(self.image.jpeg_interchange_format_length), Baseline("""4507"""))

    def test_read_rational(self):
        """Test reading RATIONAL tags and compare to known baseline values."""
        self.assertEqual(str(self.image.gps_altitude)[:13], Baseline("""2189.98969072"""))
        self.assertEqual(str(self.image.gps_latitude), Baseline("""(36.0, 3.0, 11.08)"""))
        self.assertEqual(str(self.image.gps_longitude), Baseline("""(112.0, 5.0, 4.18)"""))
        self.assertEqual(str(self.image.x_resolution), Baseline("""72.0"""))
        self.assertEqual(str(self.image.y_resolution), Baseline("""72.0"""))

    def test_read_short(self):
        """Test reading a SHORT tag and compare to a known baseline value."""
        self.assertEqual(str(self.image.metering_mode), Baseline("""5"""))

    def test_read_srational(self):
        """Test reading a SRATIONAL tag and compare to a known baseline value."""
        self.assertEqual(str(self.image.brightness_value)[:13], Baseline("""11.3644957983"""))



"""Baseline strings for test_delete_exif tests."""

from exif.tests.delete_exif_baselines.delete_ascii_tags_hex import DELETE_ASCII_TAGS_HEX_BASELINE
from exif.tests.delete_exif_baselines.delete_geotag_hex import DELETE_GEOTAG_HEX_BASELINE



"""Test behavior when opening images not conforming to EXIF specification."""

import os
import textwrap
import unittest

from baseline import Baseline

from exif import Image


NO_APP1_PNG = Baseline("""
    b'\\x89PNG\\r\\n\\x1a\\n\\x00\\x00\\x00\\rIHDR\\x00\\x00\\x00\\x02\\x00\\x00\\x00\\x02\\x08\\x06\\x00\\x00\\x00r
    \\xb6\\r$\\x00\\x00\\x00\\x01sRGB\\x00\\xae\\xce\\x1c\\xe9\\x00\\x00\\x00\\x04gAMA\\x00\\x00\\xb1\\x8f\\x0b\\xf
    ca\\x05\\x00\\x00\\x00\\tpHYs\\x00\\x00\\x0e\\xc2\\x00\\x00\\x0e\\xc2\\x01\\x15(J\\x80\\x00\\x00\\x00\\x1bIDAT
    \\x18Wcx+\\xa3\\xf2\\xdf\\xde\\xe3\\xcc\\x7f\\xc6\\xff\\x9f\\x18\\xfe+\\xef\\xf7a\\x00\\x00Q\\xe5\\x08\\x9ef\\x
    ebb3\\x00\\x00\\x00\\x00IEND\\xaeB`\\x82'
    """)


class TestInvalidFiles(unittest.TestCase):

    """Test behavior when opening images not conforming to EXIF specification."""

    def test_invalid_app1_segment(self):
        """Verify that an image with an invalid APP1 segment marker raises an IOError.

        In this case, an invalid APP1 segment marker occurs when the parser finds an 0xFFE1 in the
        image data itself (and hence finds no metadata/tags).

        """
        image_path = os.path.join(os.path.dirname(__file__), 'invalid_exif_app1.png')
        with open(image_path, 'rb') as image_file:
            my_image = Image(image_file)

        self.assertFalse(my_image.has_exif)

    def test_no_app1_segment(self):
        """Verify behavior of an image without an APP1 segment marker.

        Assert the ``has_exif`` attribute is false. Verify non-EXIF ``dir()`` list contents. Then,
        check the ``get_file()`` hexadecimal.

        """
        image_path = os.path.join(os.path.dirname(__file__), 'no_app1.png')
        with open(image_path, 'rb') as image_file:
            my_image = Image(image_file)

        self.assertFalse(my_image.has_exif)

        self.assertEqual(str(dir(my_image)), Baseline("""
            ['_segments', 'get', 'get_file', 'has_exif']
            """))

        self.assertEqual('\n'.join(textwrap.wrap(str(my_image.get_file()), 90)), NO_APP1_PNG)



"""Image EXIF metadata interface module."""

import binascii
import sys

from exif._constants import ATTRIBUTE_ID_MAP, ExifMarkers, HEX_PER_BYTE
from exif._app1_metadata import App1MetaData


class Image:

    """Image EXIF metadata interface class.

    :param file img_file: image file with EXIF metadata

    """

    has_exif = None
    """Boolean reporting whether or not the image currently has EXIF metadata."""

    def _parse_segments(self, img_hex):
        cursor = 0

        # Traverse hexadecimal string until EXIF APP1 segment found.
        while img_hex[cursor:cursor + 4] != ExifMarkers.APP1:
            cursor += HEX_PER_BYTE
            if cursor > len(img_hex):
                self.has_exif = False
                break

        self._segments['preceding'] = img_hex[:cursor]
        app1_start_index = cursor

        if self.has_exif:
            # Determine the expected length of the APP1 segment.

            app1_len = int(
                img_hex[app1_start_index + 2 * HEX_PER_BYTE:app1_start_index + 4 * HEX_PER_BYTE],
                16
            )
            cursor += app1_len * HEX_PER_BYTE

            # If the expected length stops early, keep traversing until another section is found.
            while img_hex[cursor - 2:cursor] != ExifMarkers.SEG_PREFIX:
                cursor += 2
                # raise IOError("no subsequent EXIF segment found, is this an EXIF-encoded JPEG?")
                if cursor > len(img_hex):
                    self.has_exif = False
                    break

        if self.has_exif:
            # Instantiate an APP1 segment object to create an EXIF tag interface.
            self._segments['APP1'] = App1MetaData(img_hex[app1_start_index:cursor])
            self._segments['succeeding'] = img_hex[cursor:]
        else:
            # Store the remainder of the image so that it can be reconstructed when exporting.
            self._segments['succeeding'] = img_hex[app1_start_index:]

    def __init__(self, img_file):
        self.has_exif = True
        self._segments = {}

        img_hex = binascii.hexlify(img_file.read()).upper()
        if sys.version_info[0] == 3:  # pragma: no cover
            img_hex = img_hex.decode("utf8")

        self._parse_segments(img_hex)

    def __dir__(self):
        members = ['get', 'get_file', 'has_exif', '_segments']

        if self.has_exif:
            members += self._segments['APP1'].get_tag_list()

        return members

    def __getattr__(self, item):
        return getattr(self._segments['APP1'], item)

    def __setattr__(self, key, value):
        try:
            ATTRIBUTE_ID_MAP[key]
        except KeyError:
            super(Image, self).__setattr__(key, value)
        else:
            setattr(self._segments['APP1'], key, value)

    def __delattr__(self, item):
        try:
            ATTRIBUTE_ID_MAP[item]
        except KeyError:
            super(Image, self).__delattr__(item)
        else:
            delattr(self._segments['APP1'], item)

    def __getitem__(self, item):
        return self.__getattr__(item)

    def __setitem__(self, key, value):
        self.__setattr__(key, value)

    def __delitem__(self, key):
        self.__delattr__(key)

    def delete(self, attribute):
        """Remove the specified attribute from the image.

        :param str attribute: image EXIF attribute name

        """
        self.__delattr__(attribute)

    def get(self, attribute, default=None):
        """Return the value of the specified attribute.

        If the attribute is not available or set, return the value specified by the ``default``
        keyword argument.

        :param str attribute: image EXIF attribute name
        :param default: return value if attribute does not exist
        :returns: tag value if present, ``default`` otherwise
        :rtype: corresponding Python type

        """
        try:
            retval = self.__getattr__(attribute)
        except AttributeError:
            retval = default

        return retval

    def get_file(self):
        """Generate equivalent binary file contents.

        :returns: image binary with EXIF metadata
        :rtype: str (Python 2) or bytes (Python 3)

        """
        img_hex = self._segments['preceding']

        if self.has_exif:
            img_hex += self._segments['APP1'].get_segment_hex()

        img_hex += self._segments['succeeding']

        return binascii.unhexlify(img_hex)

    def set(self, attribute, value):
        """Set the value of the specified attribute.

        :param str attribute: image EXIF attribute name
        :param value: tag value
        :type value: corresponding Python type

        """
        self.__setattr__(attribute, value)



"""Test special behavior for accessing user comment EXIF attribute."""

import os
import textwrap
import unittest

from baseline import Baseline

from exif import Image
from exif.tests.modify_exif_baselines import MODIFY_USER_COMMENT_BASELINE


class TestUserComment(unittest.TestCase):

    """Test special behavior for accessing user comment EXIF attribute."""

    def setUp(self):
        """Open sample image file in binary mode for use in test cases."""
        image_path = os.path.join(os.path.dirname(__file__), 'user_comment.jpg')
        with open(image_path, 'rb') as image_file:
            self.image = Image(image_file)

        assert self.image.has_exif

    def test_modify_ascii_same_len(self):
        """Verify that writing a same length string to an ASCII tag updates the tag."""
        # pylint: disable=protected-access
        self.image.user_comment = "Updated user comment."
        self.assertEqual(self.image.user_comment, Baseline("""Updated user comment."""))

        segment_hex = self.image._segments['APP1'].get_segment_hex()
        self.assertEqual('\n'.join(textwrap.wrap(segment_hex, 90)),
                         MODIFY_USER_COMMENT_BASELINE)

    def test_read(self):
        """Test reading a user comment."""
        self.assertEqual(
            self.image.user_comment,
            "This image shall be used in a package regression/unit test to verify reading this "
            "user comment attribute."
        )



"""Test accessing and manipulating attributes of a little endian image."""

import os
import textwrap
import unittest

from baseline import Baseline

from exif import Image
from exif.tests.little_endian_baselines import LITTLE_ENDIAN_MODIFY_BASELINE

# pylint: disable=pointless-statement, protected-access


class TestLittleEndian(unittest.TestCase):

    """Test cases for little endian images."""

    def setUp(self):
        """Open sample image file in binary mode for use in test cases."""
        little_endian_img = os.path.join(os.path.dirname(__file__), 'little_endian.jpg')
        with open(little_endian_img, 'rb') as image_file:
            self.image = Image(image_file)

        assert self.image.has_exif

    def test_modify(self):
        """Verify that modifying tags updates the tag values as expected."""
        self.image.model = "Modified"
        self.assertEqual(self.image.model, "Modified")

        self.image.gps_longitude = (12.0, 34.0, 56.789)
        self.assertEqual(str(self.image.gps_longitude), Baseline("""(12.0, 34.0, 56.789)"""))

        segment_hex = self.image._segments['APP1'].get_segment_hex()
        self.assertEqual('\n'.join(textwrap.wrap(segment_hex, 90)),
                         LITTLE_ENDIAN_MODIFY_BASELINE)

    def test_read(self):
        """Test reading tags and compare to known baseline values."""
        self.assertEqual(repr(self.image.color_space), "<ColorSpace.SRGB: 1>")
        self.assertEqual(str(self.image.datetime_original), Baseline("""2019:02:08 21:44:35"""))
        self.assertEqual(str(self.image.gps_latitude),
                         Baseline("""(79.0, 36.0, 54.804590935844615)"""))
        self.assertEqual(str(self.image.gps_longitude),
                         Baseline("""(47.0, 25.0, 34.489798675854615)"""))
        self.assertEqual(self.image.make, Baseline("""EXIF Package"""))
        self.assertEqual(self.image.model, Baseline("""Little Endian"""))
        self.assertEqual(str(self.image.resolution_unit), Baseline("""2"""))
        self.assertEqual(repr(self.image.saturation), Baseline("""<Saturation.LOW: 1>"""))
        self.assertEqual(repr(self.image.sharpness), Baseline("""<Sharpness.SOFT: 1>"""))
        self.assertEqual(str(self.image.x_resolution), Baseline("""200.0"""))
        self.assertEqual(str(self.image.y_resolution), Baseline("""200.0"""))



"""IFD SHORT tag structure parser module."""

import binascii
import struct

from exif._constants import (
    ATTRIBUTE_ID_MAP, ColorSpace, EXIF_LITTLE_ENDIAN_HEADER, Saturation, Sharpness,
    Orientation)
from exif._ifd_tag._base import Base as BaseIfdTag


class Short(BaseIfdTag):

    """IFD SHORT tag structure parser class."""

    def modify(self, value):
        """Modify tag value.

        :param value: new tag value
        :type value: corresponding Python type

        """
        # If IFD tag contains multiple values, ensure value is a tuple of appropriate length.
        if isinstance(value, tuple):
            assert len(value) == self.count
        else:
            assert self.count == 1
            value = (value,)

        for _ in range(self.count):
            if self.parent_segment_hex.endianness == EXIF_LITTLE_ENDIAN_HEADER:
                new_member_bits = struct.pack(">HH", 0, value[0])
            else:
                new_member_bits = struct.pack(">HH", value[0], 0)

            new_member_bits = binascii.hexlify(new_member_bits)
            self.parent_segment_hex.modify_hex(self.value_offset_addr, new_member_bits)
            self.value_offset = int(new_member_bits, 16)

    def read(self):
        """Read tag value.

        :returns: tag value
        :rtype: corresponding Python type

        """
        retvals = []

        for member_index in range(self.count):  # pylint: disable=unused-variable
            value_bits = struct.pack('>I', self.value_offset)
            retvals.append(struct.unpack('>HH', value_bits)[self.struct_index])

        retvals_in_enum = []  # Converted to enumerations where applicable.
        for retval in retvals:
            retvals_in_enum.append(self._short_to_enum(retval))

        if len(retvals_in_enum) == 1:
            retval = retvals_in_enum[0]
        else:
            retval = tuple(retvals_in_enum)

        return retval

    def _short_to_enum(self, read_number):
        retval = read_number

        if self.tag == ATTRIBUTE_ID_MAP["color_space"]:
            retval = ColorSpace(read_number)
        elif self.tag == ATTRIBUTE_ID_MAP["orientation"]:
            retval = Orientation(read_number)
        elif self.tag == ATTRIBUTE_ID_MAP["saturation"]:
            retval = Saturation(read_number)
        elif self.tag == ATTRIBUTE_ID_MAP["sharpness"]:
            retval = Sharpness(read_number)
        else:
            # No enumeration found, return in original form.
            pass

        return retval



"""IFD BYTE tag structure parser module."""

import struct

from exif._ifd_tag._base import Base as BaseIfdTag


class Byte(BaseIfdTag):

    """IFD BYTE tag structure parser class."""

    def modify(self, value):  # pragma: no cover
        raise NotImplementedError("this package does not yet support setting BYTE tags")

    def read(self):
        """Read tag value.

        :returns: tag value
        :rtype: corresponding Python type

        """
        retvals = []

        for member_index in range(self.count):  # pylint: disable=unused-variable
            value_bits = struct.pack('>I', self.value_offset)
            retvals.append(struct.unpack('>BBBB', value_bits)[self.struct_index])

        if len(retvals) == 1:
            retval = retvals[0]
        else:
            retval = tuple(retvals)

        return retval



"""Test special behavior for accessing Windows XP style EXIF attribute."""

import os
import unittest

from baseline import Baseline

from exif import Image


class TestWindowsXp(unittest.TestCase):

    """Test special behavior for accessing Windows XP style EXIF attribute."""

    def setUp(self):
        """Open sample image file in binary mode for use in test cases."""
        image_path = os.path.join(os.path.dirname(__file__), 'windows_xp_tags.jpg')
        with open(image_path, 'rb') as image_file:
            self.image = Image(image_file)

        assert self.image.has_exif

    def test_xp_author(self):
        """Test reading Windows XP author attribute."""
        self.assertEqual(self.image.xp_author, Baseline("""XP-Style Author"""))

    def test_xp_comment(self):
        """Test reading Windows XP comment attribute."""
        self.assertEqual(self.image.xp_comment, Baseline("""XP-Style Comment"""))

    def test_xp_keywords(self):
        """Test reading Windows XP author attribute."""
        self.assertEqual(self.image.xp_keywords, Baseline("""XP-Style Keywords"""))

    def test_xp_subject(self):
        """Test reading Windows XP author attribute."""
        self.assertEqual(self.image.xp_subject, Baseline("""XP-Style Subject"""))

    def test_xp_title(self):
        """Test reading Windows XP title attribute."""
        self.assertEqual(self.image.xp_title, Baseline("""XP-Style Title"""))



"""Test listing EXIF attributes."""

import os
import textwrap
import unittest

from baseline import Baseline

from exif import Image


class TestGetFile(unittest.TestCase):

    """Test cases for listing EXIF attributes."""

    def test_list_attributes_photo(self):
        """Verify that calling dir() on a camera photo lists the expected EXIF attributes."""
        with open(os.path.join(os.path.dirname(__file__), 'grand_canyon.jpg'), 'rb') as image_file:
            image = Image(image_file)
        dunder_dir_text = '\n'.join(textwrap.wrap(repr(sorted(dir(image))), 90))
        self.assertEqual(dunder_dir_text, Baseline("""
            ['<unknown EXIF tag 59932>', '<unknown EXIF tag 59933>', '_exif_ifd_pointer',
            '_gps_ifd_pointer', '_segments', 'aperture_value', 'brightness_value', 'color_space',
            'components_configuration', 'compression', 'datetime', 'datetime_digitized',
            'datetime_original', 'exif_version', 'exposure_bias_value', 'exposure_mode',
            'exposure_program', 'exposure_time', 'f_number', 'flash', 'flashpix_version',
            'focal_length', 'focal_length_in_35mm_film', 'get', 'get_file', 'gps_altitude',
            'gps_altitude_ref', 'gps_datestamp', 'gps_dest_bearing', 'gps_dest_bearing_ref',
            'gps_horizontal_positioning_error', 'gps_img_direction', 'gps_img_direction_ref',
            'gps_latitude', 'gps_latitude_ref', 'gps_longitude', 'gps_longitude_ref', 'gps_speed',
            'gps_speed_ref', 'gps_timestamp', 'has_exif', 'jpeg_interchange_format',
            'jpeg_interchange_format_length', 'lens_make', 'lens_model', 'lens_specification', 'make',
            'maker_note', 'metering_mode', 'model', 'orientation', 'photographic_sensitivity',
            'pixel_x_dimension', 'pixel_y_dimension', 'resolution_unit', 'scene_capture_type',
            'scene_type', 'sensing_method', 'shutter_speed_value', 'software', 'subject_area',
            'subsec_time_digitized', 'subsec_time_original', 'white_balance', 'x_resolution',
            'y_and_c_positioning', 'y_resolution']
            """))

    def test_list_attributes_simple(self):
        """Verify that calling dir() on a simple image lists the expected EXIF attributes."""
        with open(os.path.join(os.path.dirname(__file__), 'noise.jpg'), 'rb') as image_file:
            image = Image(image_file)
        dunder_dir_text = '\n'.join(textwrap.wrap(repr(sorted(dir(image))), 90))
        self.assertEqual(dunder_dir_text, Baseline("""
            ['_exif_ifd_pointer', '_segments', 'color_space', 'compression', 'datetime', 'get',
            'get_file', 'has_exif', 'jpeg_interchange_format', 'jpeg_interchange_format_length',
            'orientation', 'pixel_x_dimension', 'pixel_y_dimension', 'resolution_unit', 'software',
            'x_resolution', 'y_resolution']
            """))



import setuptools

with open("README.rst", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    python_requires='>=3.5',
    include_package_data=True,

    name='exif',
    version='0.8.4',
    author='Tyler N. Thieding',
    author_email='python@thieding.com',
    maintainer='Tyler N. Thieding',
    maintainer_email = 'python@thieding.com',
    url='https://github.com/TNThieding/exif',
    description='Read and modify image EXIF metadata using Python.',
    long_description=long_description,
    download_url='https://github.com/TNThieding/exif',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Multimedia :: Graphics :: Editors',
        ],
    license='MIT License',
    packages=setuptools.find_packages()
)



"""Hexadecimal string interface module."""

from exif._constants import EXIF_BIG_ENDIAN_HEADER, EXIF_LITTLE_ENDIAN_HEADER, HEX_PER_BYTE


class HexInterface:

    """Hexadecimal string interface class."""

    def __init__(self, hex_string):
        self.endianness = EXIF_BIG_ENDIAN_HEADER
        self._hex_string = hex_string

    @staticmethod
    def _change_endian(hexadecimal_string):
        return "".join(reversed(
            [hexadecimal_string[i:i+2] for i in range(0, len(hexadecimal_string), 2)]))

    def delete(self, offset, num_bytes):
        """Delete a sequence of hexadecimal digits from the hexadecimal string.

        :param int offset: starting byte to delete
        :param int num_bytes: number of bytes to delete

        """
        self._hex_string = (
            self._hex_string[:offset * HEX_PER_BYTE] +
            self._hex_string[(offset + num_bytes) * HEX_PER_BYTE:]
        )

    def get_hex_string(self):
        """Get the equivalent hexadecimal string.

        :returns: hexadecimal string
        :rtype: str

        """
        return self._hex_string

    def insert_null(self, offset, num_bytes):
        """Insert empty bytes (i.e. '00') into the hexadecimal string.

        :param int offset: byte to insert at
        :param int num_bytes: number of empty bytes to insert

        """
        self._hex_string = (
            self._hex_string[:offset * HEX_PER_BYTE] +
            '0' * num_bytes * HEX_PER_BYTE +
            self._hex_string[offset * HEX_PER_BYTE:]
        )

    def modify_hex(self, offset, hex_string):
        """Replace bytes in the hexadecimal string with a new string.

        :param int offset: starting byte to modify
        :param str hex_string: bytes to overwrite with

        """
        if isinstance(hex_string, bytes):
            hex_string = hex_string.decode("utf8")

        if self.endianness == EXIF_LITTLE_ENDIAN_HEADER:
            hex_string = self._change_endian(hex_string)

        self._hex_string = (
            self._hex_string[:offset * HEX_PER_BYTE] +
            hex_string +
            self._hex_string[offset * HEX_PER_BYTE + len(hex_string):]
        )

    def modify_number(self, offset, num_bytes, value):
        """Replace bytes in the hexadecimal string with a number.

        :param int offset: starting byte to modify
        :param int num_bytes: number of bytes to store value in
        :param int value: new numeric value

        """
        number_hex_bytes = hex(value).lstrip('0x').zfill(num_bytes * HEX_PER_BYTE)

        if self.endianness == EXIF_LITTLE_ENDIAN_HEADER:
            number_hex_bytes = self._change_endian(number_hex_bytes)

        self._hex_string = (
            self._hex_string[:offset * HEX_PER_BYTE] +
            number_hex_bytes +
            self._hex_string[(offset + num_bytes) * HEX_PER_BYTE:]
        )

    def read(self, offset, num_bytes):
        """Read bytes in the hexadecimal string.

        :param int offset: starting byte to read
        :param int num_bytes: number of bytes to read

        """
        hex_bytes = self._hex_string[offset * HEX_PER_BYTE:(offset + num_bytes) * HEX_PER_BYTE]

        if self.endianness == EXIF_LITTLE_ENDIAN_HEADER:
            hex_bytes = self._change_endian(hex_bytes)

        if len(hex_bytes) != num_bytes * HEX_PER_BYTE:
            raise ValueError("byte offset exceeds length of segment")

        return hex_bytes

    def set_endianness(self, endian_constant):
        """Set the endianness of the byte string.

        :param endian_constant: endian marker constant
        :type endian_constant: str (EXIF_BIG_ENDIAN_HEADER or EXIF_LITTLE_ENDIAN_HEADER)

        """
        assert endian_constant in [EXIF_BIG_ENDIAN_HEADER, EXIF_LITTLE_ENDIAN_HEADER]
        self.endianness = endian_constant

    def wipe(self, offset, num_bytes):
        """Wipe clear bytes in the hexadecimal string.

        :param int offset: starting byte to wipe
        :param int num_bytes: bytes to wipe (i.e. set to zero)

        """
        self._hex_string = (
            self._hex_string[:offset * HEX_PER_BYTE] +
            '0' * num_bytes * HEX_PER_BYTE +
            self._hex_string[(offset + num_bytes) * HEX_PER_BYTE:]
        )



"""IFD RATIONAL tag structure parser module."""

import binascii
import struct
from fractions import Fraction

from exif._constants import EXIF_LITTLE_ENDIAN_HEADER
from exif._ifd_tag._base import Base as BaseIfdTag


class Rational(BaseIfdTag):

    """IFD RATIONAL tag structure parser class."""

    def modify(self, value):
        """Modify tag value.

        :param value: new tag value
        :type value: corresponding Python type

        """
        # If IFD tag contains multiple values, ensure value is a tuple of appropriate length.
        if isinstance(value, tuple):
            assert len(value) == self.count
        else:
            assert self.count == 1
            value = (value,)

        cursor = 0xA + self.value_offset
        for member_index in range(self.count):
            fraction = Fraction(value[member_index]).limit_denominator()

            if self.parent_segment_hex.endianness == EXIF_LITTLE_ENDIAN_HEADER:
                new_member_bits = struct.pack(
                    ">LL", fraction.denominator, fraction.numerator)
            else:
                new_member_bits = struct.pack(
                    ">LL", fraction.numerator, fraction.denominator)

            new_member_hex = binascii.hexlify(new_member_bits)
            self.parent_segment_hex.modify_hex(cursor, new_member_hex)
            cursor += 8

    def read(self):
        """Read tag value.

        :returns: tag value
        :rtype: corresponding Python type

        """
        retvals = []
        cursor = 0xA + self.value_offset

        for member_index in range(self.count):  # pylint: disable=unused-variable
            data_format = ">LL"
            numerator, denominator = struct.unpack(data_format, binascii.unhexlify(
                self.parent_segment_hex.read(cursor, 8)))
            cursor += 8
            if self.parent_segment_hex.endianness == EXIF_LITTLE_ENDIAN_HEADER:
                retvals.append(float(denominator) / float(numerator))
            else:
                retvals.append(float(numerator) / float(denominator))

        if len(retvals) == 1:
            retval = retvals[0]
        else:
            retval = tuple(retvals)

        return retval



"""Baseline strings for test_modify_exif tests."""

from exif.tests.modify_exif_baselines.modify_ascii_same_len_hex import (
    MODIFY_ASCII_SAME_LEN_HEX_BASELINE)
from exif.tests.modify_exif_baselines.modify_ascii_shorter_hex import (
    MODIFY_ASCII_SHORTER_HEX_BASELINE)
from exif.tests.modify_exif_baselines.modify_orientation_hex import ROTATED_GRAND_CANYON_HEX
from exif.tests.modify_exif_baselines.modify_rational_hex import MODIFY_RATIONAL_HEX_BASELINE
from exif.tests.modify_exif_baselines.modify_srational_hex import MODIFY_SRATIONAL_HEX_BASELINE
from exif.tests.modify_exif_baselines.modify_user_comment import MODIFY_USER_COMMENT_BASELINE



"""Legacy Windows XP style tag structure parser module."""

import warnings

from exif._ifd_tag._base import Base as BaseIfdTag


class WindowsXp(BaseIfdTag):

    """Legacy Windows XP style tag structure parser class."""

    def modify(self, value):  # pragma: no cover
        """Modify tag value.

        :param value: new tag value
        :type value: corresponding Python type

        """
        raise NotImplementedError("this package does not yet support setting Windows XP style tags")

    def read(self):
        """Read tag value.

        In string types, the count refers to how many characters exist in the string.

        :returns: tag value
        :rtype: corresponding Python type

        """
        retval_chars = []
        cursor = 0xA + self.value_offset

        for member_index in range(self.count - 1):  # pylint: disable=unused-variable
            current_ascii_val = self.parent_segment_hex.read(cursor, 1)

            if not current_ascii_val:  # pragma: no cover
                warnings.warn("reached end of string prematurely", RuntimeWarning)
                break

            if current_ascii_val == "00":
                break

            retval_chars.append(chr(int(current_ascii_val, 16)))
            cursor += 2

        return ''.join(retval_chars)
